#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <netdb.h>
#endif

#include "tig.h"

#ifdef BIG_ENDIAN
static int16_t swap16(int16_t v)
{
	return ((v >> 8) & 0x00FF) |
	       ((v << 8) & 0xFF00);
}

static int32_t swap32(int32_t v)
{
	return ((v >> 24) & 0x000000FF) |
	       ((v >> 8)  & 0x0000FF00) |
	       ((v << 8)  & 0x00FF0000) |
	       ((v << 24) & 0xFF000000);
}

static int64_t swap64(int64_t v)
{
	return ((v >> 56) & 0x00000000000000FFL) |
	       ((v >> 40) & 0x000000000000FF00L) |
	       ((v >> 24) & 0x0000000000FF0000L) |
	       ((v >> 8)  & 0x00000000FF000000L) |
	       ((v << 8)  & 0x000000FF00000000L) |
	       ((v << 24) & 0x0000FF0000000000L) |
	       ((v << 40) & 0x00FF000000000000L) |
	       ((v << 56) & 0xFF00000000000000L);
}
#endif

int16_t to_little_16(int16_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap16(v);
#endif
}

int32_t to_little_32(int32_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap32(v);
#endif
}

int64_t to_little_64(int64_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap64(v);
#endif
}

int16_t from_little_16(int16_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap16(v);
#endif
}

int32_t from_little_32(int32_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap32(v);
#endif
}

int64_t from_little_64(int64_t v)
{
#ifndef BIG_ENDIAN
	return v;
#else
	return swap64(v);
#endif
}

static const int MAX_FRAME_SIZE = 1000000;

struct con_t *con_init(const char *host, int port)
{
    char sport[6];
    sprintf(sport, "%d", port);
    struct con_t *con = malloc(sizeof(struct con_t));
    int status;
    struct addrinfo hints, *res;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = 0;
    status = getaddrinfo(host, sport, &hints, &res);
    if(status != 0) return NULL;
    con->sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if(con->sd < 0) return NULL;
    status = connect(con->sd, res->ai_addr, res->ai_addrlen);
    if(status != 0) return NULL;
    freeaddrinfo(res);
    return con;
}

void con_prep(struct con_t *con)
{
    con->buf = malloc(MAX_FRAME_SIZE);
    con->bufix = 0;
    int32_t temp = 0;
    con_put_int32_t(con, temp);
}

static void con_put_anything(struct con_t *con, const void *data, int datasiz)
{
    memcpy(con->buf + con->bufix, data, datasiz);
    con->bufix += datasiz;
}

void con_put_int32_t(struct con_t *con, int32_t v)
{
    con_put_anything(con, &v, 4);
}

void con_put_charptr(struct con_t *con, const char* v)
{
    int len = strlen(v);
    con_put_anything(con, &len, 2);
    con_put_anything(con, v, len);
}

void con_put_block(struct con_t *con, void* v, int len)
{
    con_put_anything(con, &len, 2);
    con_put_anything(con, v, len);
}

void con_send(struct con_t *con)
{
    int wrlen = con->bufix - 4;
    memcpy(con->buf, &wrlen, 4);
    send(con->sd, con->buf, con->bufix, 0);
    free(con->buf);
}

void con_recv(struct con_t *con)
{
    con_get_int32_t(con);
}

static void con_get_anything(struct con_t *con, void *buf, int siz)
{
    int ix = 0;
    while(ix < siz)
    {
        ix += recv(con->sd, buf + ix, siz - ix, 0);
    }
}

int32_t con_get_int32_t(struct con_t *con)
{
    int32_t res;
    con_get_anything(con, &res, 4);
    return res;
}

char* con_get_charptr(struct con_t *con)
{
    int16_t len;
    con_get_anything(con, &len, 2);
    char* buf  = malloc(len + 1);
    con_get_anything(con, buf, len);
    buf[len] = 0;
    return buf;
}

void* con_get_block(struct con_t *con)
{
    int16_t len;
    con_get_anything(con, &len, 2);
    char* buf  = malloc(len);
    con_get_anything(con, buf, len);
    return buf;
}

void con_close(struct con_t *con)
{
#ifdef WIN32
    closesocket(con->sd);
#else
    close(con->sd);
#endif
}
